# -*- coding: utf-8 -*-
# Copyright © 2023-present Worldr Technologies Limited. All Rights Reserved.
"""Console entry point."""
from __future__ import annotations

import logging
import logging.config
import os
import sys
from pathlib import Path

import click
import pendulum
import structlog
from click_help_colors import HelpColorsCommand  # type: ignore[import]
from dotenv import load_dotenv
from orjson import loads
from redmail import EmailSender, gmail  # pyright: ignore [reportPrivateImportUsage]
from rich.console import Console
from rich.prompt import Confirm
from rich.traceback import install

from nhound import __version__
from nhound.email import IEMail
from nhound.inotion import INotion, INotionError
from nhound.utils import COLOUR_INFO, VersionCheck, check_if_latest_version, wprint

# Rich.
install(show_locals=True)

EXIT_CODE_SUCCESS = 0
EXIT_CODE_OPERATION_FAILED = 1
EXIT_CODE_NOTION_API_FAILED = 2


pre_chain = [
    # Add the log level and a timestamp to the event_dict if the log entry
    # is not from structlog.
    structlog.stdlib.add_log_level,
    # Add extra attributes of LogRecord objects to the event dictionary
    # so that values passed in the extra parameter of log methods pass
    # through to log output.
    structlog.stdlib.ExtraAdder(),
]


def configure_logging(log_level: str, verbose: bool) -> None:
    """Configure all the logging."""
    # Logging levels
    # https://www.structlog.org/en/stable/_modules/structlog/_log_levels.html?highlight=log%20level
    _lvl = {
        "critical": 50,
        "error": 40,
        "warning": 30,
        "info": 20,
        "debug": 10,
        "notset": 0,
    }

    # Structlog processors. Order appears to matter…
    shared_processors = [
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.processors.StackInfoRenderer(),
        structlog.processors.TimeStamper(fmt="iso"),
        structlog.processors.format_exc_info,
        structlog.processors.CallsiteParameterAdder(
            [
                structlog.processors.CallsiteParameter.FILENAME,
                structlog.processors.CallsiteParameter.FUNC_NAME,
                structlog.processors.CallsiteParameter.LINENO,
            ]
        ),
    ]

    class VerboseFilter(logging.Filter):
        """Filter log entries on verbose flag."""

        def __init__(self, param: str = "") -> None:
            """Initialise."""
            self.param = param
            super()

        def filter(self, _: logging.LogRecord) -> bool:  # noqa: A003
            # We have no choice in the method's name.
            # We do not care about record thus mark it as _.
            #
            # The pragam no cover can be removed once we actually use
            # something useful.
            return verbose  # pragma: no cover

    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": True,  # tabula raza.
            "filters": {
                "myfilter": {
                    "()": VerboseFilter,
                    "param": "noshow",
                }
            },
            "formatters": {
                "plain": {
                    "()": structlog.stdlib.ProcessorFormatter,
                    "processors": [
                        *shared_processors,
                        structlog.stdlib.ProcessorFormatter.remove_processors_meta,
                        structlog.processors.JSONRenderer(),
                    ],
                    "foreign_pre_chain": pre_chain,
                },
                "colored": {
                    "()": structlog.stdlib.ProcessorFormatter,
                    "processors": [
                        *shared_processors,
                        structlog.stdlib.ProcessorFormatter.remove_processors_meta,
                        structlog.dev.ConsoleRenderer(colors=True),
                    ],
                    "foreign_pre_chain": pre_chain,
                },
            },
            "handlers": {
                "default": {
                    "level": _lvl[log_level],
                    "class": "logging.StreamHandler",
                    "filters": ["myfilter"],
                    "formatter": "colored",
                },
                "file": {
                    "level": _lvl[log_level],
                    "class": "logging.handlers.WatchedFileHandler",
                    "filename": "nhound.log",
                    "formatter": "plain",
                },
            },
            # Define all the loggers you want!
            "loggers": {
                "nhound": {
                    "handlers": ["default", "file"],
                    "level": _lvl[log_level],
                    "propagate": True,
                },
                "requests": {
                    "handlers": ["default", "file"],
                    "level": _lvl[log_level],
                    "propagate": True,
                },
                "rich": {
                    "handlers": ["default", "file"],
                    "level": _lvl[log_level],
                    "propagate": True,
                },
            },
        }
    )
    structlog.configure(
        processors=[
            *shared_processors,  # type: ignore[list-item]
            structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
        ],
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.make_filtering_bound_logger(_lvl[log_level]),
        cache_logger_on_first_use=True,
    )


@click.command(
    cls=HelpColorsCommand,
    help_headers_color="blue",
    help_options_color="magenta",
)
@click.option(
    "-l",
    "--log-level",
    default="info",
    show_default=True,
    type=click.Choice(
        ["notset", "debug", "info", "warning", "error", "critical"],
        case_sensitive=False,
    ),
    help="Chose the logging level from the available options. "
    "This affect the file logs as well.",
)
@click.option(
    "-e",
    "--env",
    default=Path(Path.cwd() / ".env"),
    show_default=True,
    type=click.Path(exists=True),
    help="Which .env file to load.",
)
@click.option("-v", "--version", is_flag=True, help="Print the version and exit")
@click.option("--verbose", is_flag=True, help="Print the logs to stdout")
def main(
    log_level: str,
    env: Path,
    version: bool,
    verbose: bool,
) -> None:
    """Setupr ships the Worldr infrastructure.

    Note that <semver> must be a valid semantic version. This is
    different for all the scripts. Please check the user documentation
    for the exact values.
    """
    # Start time.
    start_time = pendulum.now("UTC")

    # Prints the current version and exits.
    if version:
        click.echo(__version__)
        sys.exit(EXIT_CODE_SUCCESS)

    # Configure logging.
    configure_logging(log_level, verbose)
    rlog = structlog.get_logger("nhound")
    rlog.debug(
        "All the loggers",
        loggers=list(logging.root.manager.loggerDict),
    )

    # Configure the console.
    console = Console()
    console.rule(f"[{COLOUR_INFO}]Notion Hound bot")

    # Check latest version.
    _version_check()

    # Do all the hard work.
    rlog.debug("Starting real work…")
    status = _do_stuff(rlog, env)

    # We should be done…
    if status:
        wprint("Operation was successful.", level="success")
    else:
        wprint("Operation might have failed.", level="warning")
    rlog.info(
        "That's all folks!",
        duration=(
            pendulum.now("UTC") - start_time
        ).in_words(),  # pyright: ignore[reportGeneralTypeIssues]
    )
    sys.exit(EXIT_CODE_SUCCESS)


def _do_stuff(rlog: structlog.BoundLogger, env: Path) -> bool:  # pragma: no cover
    """Do stuff.

    Why not unit tests? Well, this is actually doing work. We could mock
    everything, but is there a point to doing that?

    A functional tests might be better. Again, it would not be trivial
    to set up.
    """
    # Get enviorment variables from .env file.
    rlog.debug("Loading environment variables from .env file.", env=env)
    load_dotenv(env)  # take environment variables from .env.
    token = ""  # There should never be a real value here.  # nosec
    try:
        token = os.environ["NHOUND_NOTION_TOKEN"]
    except KeyError as e:
        rlog.exception("Missing environment variable", var=e)
        wprint("Missing environment variable NHOUND_NOTION_TOKEN.", level="error")
        sys.exit(EXIT_CODE_OPERATION_FAILED)

    # Get UUID of Notion pages from environment variable.
    uuids = loads(os.environ["NHOUND_PAGES_UUIDS"])

    # Set up email forwarding.
    #
    # Note that if NHOUND_SMTP_USERNAME and NHOUND_SMTP_PASSWORD are not set,
    # then the email will be sent without authentication.
    my_sender = EmailSender(
        host=os.environ["NHOUND_SMTP_HOST"],
        port=int(os.environ["NHOUND_SMTP_PORT"]),
        use_starttls=os.environ["NHOUND_SMTP_USE_STARTTLS"].lower().capitalize()
        is False,
    )
    if os.getenv("NHOUND_SMTP_USERNAME", None):
        rlog.info("Using Google SMTP server.")
        gmail.username = os.environ["NHOUND_SMTP_USERNAME"]
        gmail.password = os.environ["NHOUND_SMTP_PASSWORD"]
        my_sender = gmail
    else:
        rlog.warning("Using custom SMTP server. Probably testing…")
        rlog.warning("Run: `python -m smtpd -n -c DebuggingServer localhost:1025`")

    email = IEMail(my_sender, os.environ["NHOUND_SMTP_EMAIL_SENDER"])

    # Do stuff with Notion API.
    status = True
    try:
        inotion = INotion(
            token,
            int(os.getenv("NHOUND_PAGES_ARE_STALE_AFTER_X_WEEKS", 13)),
        )
        for data in inotion.get_email_data(uuids):
            status = status & email.send(
                receivers=[
                    data[0].email,
                ],
                body_params={"name": data[0].name, "pages": data[1]},
            )
    except INotionError as e:
        rlog.exception("INotionError", error=e)
        sys.exit(EXIT_CODE_NOTION_API_FAILED)

    if not status:
        wprint("Email sending failed.", level="warning")
        return False
    return True


def _version_check() -> None:
    """Check if we are running the latest verion from GitHub."""
    check = check_if_latest_version()
    if check == VersionCheck.LATEST:
        wprint(f"This is the latest version {__version__}.", level="info")
    elif check == VersionCheck.LAGGING:
        wprint("there is a new version available: please update.", level="warning")
        if Confirm.ask("Exit and update?", default=True):
            wprint(
                "Please run [i]python -m pip install -U nhound[/i]",
                level="info",
            )
            sys.exit(EXIT_CODE_SUCCESS)
        wprint("Proceeding with old version…", level="warning")
    elif check == VersionCheck.UNKNOWN:
        wprint("Could not check for newer versons.", level="warning")
    else:  # pragma: no cover
        # This should never, ever happen!
        wprint("This is bug, please report!", level="error")


if __name__ == "__main__":  # pragma: no cover
    main()



# -*- coding: utf-8 -*-
# Copyright © 2023-present Worldr Technologies Limited. All Rights Reserved.
"""Utilities."""
import enum
from typing import TYPE_CHECKING

from rich.console import Console

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Sequence

from typing import Any

import requests
import structlog

from nhound import __version__

# https://www.nordtheme.com/docs/colors-and-palettes
COLOUR_FAIL = "#bf616a"
COLOUR_INFO = "#5e81ac"
COLOUR_NOTE = "#81a1c1"
COLOUR_SUCC = "#a3be8c"
COLOUR_WARN = "#d08770"
COLOUR_GREY = "#777777"

GITHUB_URL = "https://api.github.com/repos/worldr/nhound/releases/latest"
rlog = structlog.get_logger("nhound.utils")


class VersionCheck(enum.Enum):
    """Version check ENUM."""

    LATEST = enum.auto()
    LAGGING = enum.auto()
    UNKNOWN = enum.auto()


def join_with_oxford_commas(obj_list: "Sequence[Any]") -> str:
    """Oxford commas for lists.

    Takes a list of objects and returns their string representations,
    separated by commas and with 'and' between the penultimate and final
    items
    """
    if not obj_list:
        return ""
    size = len(obj_list)
    if size == 1:
        return f"{obj_list[0]!s}"
    return (
        ", ".join(str(obj) for obj in obj_list[: size - 1])
        + f", and {obj_list[size - 1]!s}"
    )


def check_if_latest_version() -> VersionCheck:
    """Check if there is a new version published on GitHub."""
    response = requests.get(GITHUB_URL, timeout=60)  # A minute time out.
    if response.status_code == 200:
        latest_version = response.json()["tag_name"]
        if latest_version == f"v{__version__}":
            return VersionCheck.LATEST
        return VersionCheck.LAGGING
    return VersionCheck.UNKNOWN


def wprint(text: str, level: str = "") -> None:
    """Print wrapper.

    If there is no level, we just print it with whatever markup.
    """
    console = Console()
    if level == "note":
        console.print(f"   {text}", style=COLOUR_NOTE)
    elif level == "info":
        console.print(f":thumbs_up: {text}", style=COLOUR_INFO)
    elif level == "warning":
        console.print(f":warning-emoji:  {text}", style=COLOUR_WARN)
    elif level == "success":
        console.print(f":heavy_check_mark:  {text}", style=COLOUR_SUCC)
    elif level == "failure":
        console.print(f":x: {text}", style=COLOUR_FAIL)
    else:
        console.print(f"   {text}")


if __name__ == "__main__":  # pragma: no cover
    # We do not need to test any of this… It is just an example!
    # >>> python nhound/print.py
    TEXT = (
        "[i]I Am Malenia[/i], "
        "[b]Blade Of Miquella[/b], And "
        "[i][b]I Have Never Known Defeat.[/i][/b] "
        "[black]:skull:[/black]"
    )
    wprint(f"{TEXT}")
    wprint(f"{TEXT}", "note")
    wprint(f"{TEXT}", "info")
    wprint(f"{TEXT}", "success")
    wprint(f"{TEXT}", "warning")
    wprint(f"{TEXT}", "failure")



# -*- coding: utf-8 -*-
# Copyright © 2023-present Worldr Technologies Limited. All Rights Reserved.
"""Dehumanize a date string."""
import string
from typing import Any

import structlog

from nhound import NOW

rlog = structlog.get_logger("nhound.dehumanize")


def dehumanize(text: str) -> Any:
    """Dehumanize a date string."""
    # Clean up the input text.
    text = text.lower().translate(str.maketrans("", "", string.punctuation))

    # Try to split into value and unit.
    try:
        span, unit = text.split(" ")
    except ValueError as e:
        rlog.exception(e)
        rlog.error("Failed to parse humanized text", text=text)
        return NOW

    # Value must be an interger.
    try:
        value = int(span)
    except ValueError as e:
        if span == "a":
            value = 1
        else:
            rlog.exception(e)
            rlog.error("Failed to cast span of humanized text", text=text, span=span)
            return NOW

    # Parse the unit.
    if unit in ["day", "days"]:
        return NOW.subtract(days=int(value))
    if unit in ["week", "weeks"]:
        return NOW.subtract(weeks=int(value))
    if unit in ["month", "months"]:
        return NOW.subtract(months=int(value))
    if unit in ["year", "years"]:
        return NOW.subtract(years=int(value))

    # Default is now.
    return NOW



# -*- coding: utf-8 -*-
# Copyright © 2023-present Worldr Technologies Limited. All Rights Reserved.
"""Nhound."""
from importlib import metadata

import pendulum

__version__ = metadata.version(__package__)

del metadata  # optional, avoids polluting the results of dir(__package__)

NOW = pendulum.now("UTC")  # We should have a basic timer of now.



# -*- coding: utf-8 -*-
# Copyright © 2023-present Worldr Technologies Limited. All Rights Reserved.
"""Simple Notion user model."""
from collections import namedtuple

import structlog

rlog = structlog.get_logger("nhound.user")

Page = namedtuple(
    "Page",
    [
        "uuid",
        "title",
        "url",
        "created_time",
        "last_edited_time",
        "threashold_time",
    ],
)


class User:
    """A Notion user model."""

    def __init__(self, uuid: str, name: str, email: str) -> None:
        """Init."""
        self._uuid = uuid
        self._name = name
        self._email = email
        self.pages: set[Page] = set()

    def __repr__(self) -> str:
        """Repr."""
        return (
            f"User("
            f"uuid={self.uuid}, "
            f"name={self.name}, "
            f"email={self.email}, "
            ")"
        )

    @property
    def uuid(self) -> str:
        """Get uuid."""
        return self._uuid

    @property
    def name(self) -> str:
        """Get name."""
        return self._name

    @property
    def email(self) -> str:
        """Get email."""
        return self._email



# -*- coding: utf-8 -*-
# Copyright © 2023-present Worldr Technologies Limited. All Rights Reserved.id
"""Email sending module."""
import structlog
from redmail import EmailSender  # pyright: ignore [reportPrivateImportUsage]

from nhound.user import Page

rlog = structlog.get_logger("nhound.email")


class IEMail:
    """Interface for email."""

    # Template for email body in HTML format.
    html = """
<h1>Hi {{ name }},</h1>
<p>The following {{pages|length}} Notion page{% if pages|length > 1 %}s{% endif %} require your attention:</p>
<dl>
{% for page in pages %}
    <dt>{{ page.title }}</dt>
    <dd>{{ page.url }}</dd>
{% endfor %}
</dl>
<p>Either update them or archive them!</p>

<p>Thank you.</p>

<p>Nhound bot.</p>
    """

    # Template for email body in text format.
    text = """
Hi {{ name }},

The following {{pages|length}} Notion pages{% if pages|length > 1 %}s{% endif %}require your attention:
{% for page in pages %}
    - {{ page.title }} -- {{ page.url }}.
{% endfor %}

Either update them or archive them!

Thank you.

Nhound bot.
{{ company }}
    """

    def __init__(self, email: EmailSender, sender: str) -> None:
        """Initialize."""
        self._email = email
        self._sender = sender

    def send(self, receivers: list, body_params: dict[str, str | list[Page]]) -> bool:
        """Send email."""
        sz = len(body_params["pages"])
        if sz == 0:
            return True
        subject = f"{sz} Notion pages requires your attention"
        if sz == 1:
            subject = f"{sz} Notion page requires your attention"
        try:
            with self._email:
                self._email.send(
                    subject=subject,
                    sender=self._sender,
                    receivers=receivers,
                    text=self.text,
                    html=self.html,
                    body_params=body_params,
                )
        except ConnectionRefusedError as e:
            rlog.error(
                "Failed to send email",
                error=e,
                host=self._email.host,
                port=self._email.port,
                receivers=receivers,
            )
            return False
        else:
            return True



# -*- coding: utf-8 -*-
# Copyright © 2023-present Worldr Technologies Limited. All Rights Reserved.
"""A cohort of Notion users."""
import os
from contextlib import suppress

import pendulum
import structlog
from rich import print as rprint

from nhound import NOW
from nhound.user import Page, User
from nhound.utils import wprint

rlog = structlog.get_logger("nhound.cohort")


class Cohort:
    """A collection of users."""

    def __init__(self) -> None:
        """Init.

        We care just the day, we care not about specific times.
        """
        self._users = {}  # type: dict[str, User]
        self.now = pendulum.datetime(NOW.year, NOW.month, NOW.day)
        _interval = 13  # Default to 13 weeks, or 3 months, ish.
        with suppress(ValueError):
            _interval = int(
                os.environ.get("NHOUND_PAGES_ARE_STALE_AFTER_X_WEEKS", f"{_interval}")
            )
        self.stale = NOW.subtract(weeks=_interval)

    @property
    def size(self) -> int:
        """Get size."""
        return len(self._users)

    def add_user(self, user: User) -> None:
        """Add user."""
        if user.uuid in self._users:
            msg = f"User with uuid {user.uuid} already exists"
            rlog.warning(msg)
            raise ValueError(msg)
        self._users[user.uuid] = user

    def get_by_uuid(self, uuid: str) -> User | None:
        """Get user by its uuid.

        This is unique.
        """
        try:
            return self._users[uuid]
        except KeyError:
            msg = f"User with uuid {uuid} not found"
            rlog.warning(msg)
            return None

    def get_by_name(self, name: str) -> tuple[User, ...]:
        """Get user by thier name.

        This is not unique.
        """
        results = []
        for _, user in self._users.items():
            if user.name == name:
                results.append(user)
        return tuple(results)

    def get_by_email(self, email: str) -> tuple[User, ...]:
        """Get user by thier email.

        This is not unique.
        """
        results = []
        for _, user in self._users.items():
            if user.email == email:
                results.append(user)
        return tuple(results)

    def print_data(self) -> None:  # pragma: no cover
        """Bleurgh.

        There is no unit test here since, in all likelyhood, this will
        never be used as is.
        """
        for uuid, user in self._users.items():
            if user.pages:
                rprint(f"{uuid} {user.name} → ")
                for page in user.pages:
                    if page.last_edited_time < page.threashold_time:
                        wprint(
                            f"{page.title} is stale, "
                            f"it was editted {page.last_edited_time.diff_for_humans(NOW)} now. "
                            f"<{page.url}>",
                            level="warning",
                        )
                    else:
                        wprint(f"{page.title} is fresh.", level="info")

    def get_data_for_email(self) -> list[tuple[User, list[Page]]]:
        """Get all the data in a format email can understand."""
        ret = []
        for _, user in self._users.items():
            pages = [x for x in user.pages if x.last_edited_time < x.threashold_time]
            if pages:
                ret.append((user, pages))
        return ret



# -*- coding: utf-8 -*-
# Copyright © 2023-present Worldr Technologies Limited. All Rights Reserved.
"""Interface to Notion."""
import logging
import typing
from re import search

import pendulum
import structlog
from notion_client import APIResponseError, Client
from notion_client.helpers import collect_paginated_api, is_full_page
from pendulum.datetime import DateTime

from nhound import NOW
from nhound.cohort import Cohort
from nhound.dehumanize import dehumanize
from nhound.user import Page, User

rlog = structlog.get_logger("nhound.inotion")


class INotionError(Exception):
    """Base class for Notion errors."""


class INotion:
    """A interface to Notion's API."""

    _nhound_delimiters: typing.ClassVar[str] = "nhound{(.+?)}"

    def __init__(self, token: str, threashold: int = 13) -> None:
        """Init."""
        logger = structlog.wrap_logger(
            logging.getLogger("notion-client"),
            logger_factory=structlog.stdlib.LoggerFactory(),
            wrapper_class=structlog.stdlib.BoundLogger,
        )
        self._notion = Client(auth=token, logger=logger, log_level=logging.DEBUG)
        self._cohort = Cohort()
        self._nhound_default_threashold = threashold
        rlog.info("Initialized INotion", threashold=self._nhound_default_threashold)

    def get_users(self) -> None:
        """Get users."""
        rlog.debug("get_users")
        try:
            list_users_response = self._notion.users.list()
        except APIResponseError as e:
            rlog.exception(e)
            msg = "Failed to get users from Notion."
            rlog.error(msg)
            raise INotionError(msg) from e
        for item in list_users_response["results"]:  # type: ignore[index]
            if item["type"] == "person":
                self._cohort.add_user(
                    User(item["id"], item["name"], item["person"]["email"])
                )
        rlog.info("Got users from Notion", count=self._cohort.size)

    def _parse_callout_block(
        self, blocks: list[typing.Any]
    ) -> tuple[list[User], DateTime]:
        """Analyse any callout block."""
        users: list[User] = []
        threashold = NOW.subtract(weeks=self._nhound_default_threashold)
        for block in blocks:
            if block["type"] == "callout":
                for item in block["callout"]["rich_text"]:
                    if item["type"] == "mention":
                        if "user" not in item["mention"]:
                            rlog.debug("No user in callout", item=item)
                            continue
                        usr = self._cohort.get_by_uuid(item["mention"]["user"]["id"])
                        if usr is not None:
                            users.append(usr)
                        rlog.debug("Found callout user", user=usr)
                    if item["type"] == "text":
                        try:
                            # This will overwirght the standard threashold.
                            date = search(  # type: ignore [union-attr]
                                self._nhound_delimiters, item["text"]["content"]
                            ).group(  # pyright: ignore [reportOptionalMemberAccess]
                                1
                            )
                            threashold = dehumanize(date)
                            rlog.debug(
                                "Found nhoud callout date",
                                text=item["text"]["content"],
                                date=date,
                                threashold=threashold,
                            )
                        except AttributeError:
                            continue
        return (users, threashold)

    def _get_page_data(self, _id: str) -> None:
        page = self._notion.pages.retrieve(_id)
        title = "UNSET"
        try:
            title = page["url"].rsplit("/", 1)[-1].rsplit("-", 1)[0]  # type: ignore[index]
        except KeyError as e:
            rlog.exception(e)
        blocks = self._notion.blocks.children.list(_id)["results"]  # type: ignore[index]
        users, threashold = self._parse_callout_block(blocks)
        my_page = Page(
            _id,
            title,
            page.get("url"),  # type: ignore[union-attr]
            pendulum.parse(page.get("created_time")),  # type: ignore[union-attr]
            pendulum.parse(page.get("last_edited_time")),  # type: ignore[union-attr]
            threashold,
        )
        if not users:
            # We have users in the callout block.
            for user in users:
                user.pages.add(my_page)  # pyright: ignore [reportOptionalMemberAccess]
        else:
            # Wehave no users in the callout block.
            usr = self._cohort.get_by_uuid(page.get("created_by")["id"])  # type: ignore[union-attr]
            if usr is not None:
                usr.pages.add(my_page)
            usr = self._cohort.get_by_uuid(page.get("last_edited_by")["id"])  # type: ignore[union-attr]
            if usr is not None:
                usr.pages.add(my_page)

        for block in blocks:
            if block["type"] == "child_page":
                self._get_page_data(block["id"])
            if (
                block["type"] == "child_database"
                and "meeting" not in block.get("child_database").get("title").lower()
            ):
                self._get_database_data(block["id"])

    def _get_database_data(self, _id: str) -> None:
        """Get database data."""
        full_or_partial_pages = collect_paginated_api(
            self._notion.databases.query, database_id=_id
        )
        for page in full_or_partial_pages:
            if not is_full_page(page):
                continue
            title = "UNSET"
            try:
                title = page["url"].rsplit("/", 1)[-1].rsplit("-", 1)[0]
            except KeyError as e:
                rlog.exception(e)
            _tmp = Page(
                _id,
                title,
                page.get("url"),
                pendulum.parse(  # pyright: ignore [reportPrivateImportUsage]
                    page.get("created_time")
                ),
                pendulum.parse(  # pyright: ignore [reportPrivateImportUsage]
                    page.get("last_edited_time")
                ),  # pyright: ignore [reportPrivateImportUsage]
                NOW.subtract(weeks=self._nhound_default_threashold),
            )
            usr = self._cohort.get_by_uuid(page.get("created_by")["id"])
            if usr is not None:
                usr.pages.add(_tmp)
            usr = self._cohort.get_by_uuid(page.get("last_edited_by")["id"])
            if usr is not None:
                usr.pages.add(_tmp)
            rlog.info("Found a database entry", page=_tmp)

    def stuff(self, uuids: tuple[typing.Any, ...]) -> None:
        """Stuff."""
        rlog.debug("stuff start")
        self._get_users()
        self._get_pages(uuids)
        self._cohort.print_data()
        rlog.debug("stuff end")

    def _get_users(self) -> None:
        """Get all the Notion users."""
        try:
            self.get_users()
        except APIResponseError as e:
            msg = "Failed to get pages from Notion."
            rlog.error(msg)
            raise INotionError(msg) from e

    def _get_pages(self, uuids: tuple[typing.Any, ...]) -> None:
        """Get all the Notion pages."""
        for _id in uuids:
            try:
                self._get_page_data(_id)
            except APIResponseError as e:
                msg = "Failed to get page from Notion."
                rlog.error(msg, uuid=_id, error=e)
                continue

    def get_email_data(
        self, uuids: tuple[typing.Any, ...]
    ) -> list[tuple[User, list[Page]]]:
        """Get data suitable for sending emails."""
        rlog.debug("stuff start")
        self._get_users()
        self._get_pages(uuids)
        self._cohort.print_data()
        return self._cohort.get_data_for_email()
